@@ -99,6 +99,7 @@ sub new {
     }
 
     if ($params{pid_file}) {
+        $self->{pid_file} = $params{pid_file};
         $self->{pid} = System::Process::pidinfo(
             file    =>  $params{pid_file}
         );
@@ -119,6 +120,12 @@ sub new {
     if ($params{group}) {
         $self->{_args}->{group} = $params{group};
     }
+    if ($params{progress_log}) {
+        $self->{_args}->{progress_log} = $params{progress_log};
+    }
+    if ($params{grace_restart_file}) {
+        $self->{_args}->{grace_restart_file} = $params{grace_restart_file};
+    }
 
     bless $self, $class;
     return $self;
@@ -142,8 +149,9 @@ sub run {
         return 1;
     }
 
-    if ($self->can($ARGV[0])) {
-        my $sub = $ARGV[0];
+    my $sub = $ARGV[0];
+    $sub =~ s/-/_/g;
+    if ($self->can($sub)) {
         $self->$sub();
     }
     else {
@@ -152,10 +160,8 @@ sub run {
     return 1;
 }
 
-
 sub start {
     my $self = shift;
-    
     print STARTING;
     $self->before_start();
     # TODO: Add command check
@@ -165,8 +171,10 @@ sub start {
         return;
     }
 
+    unlink $self->{_args}->{grace_restart_file} if $self->{_args}->{grace_restart_file};
+
     system(@command);
-    
+
     my $code = $?;
     $code = $code >> 8;
     $self->after_start();
@@ -182,10 +190,10 @@ sub start {
 }
 
 
-sub stop {
-    my $self = shift;
+sub _stop {
+    my ($self, $mode) = @_;
 
-    unless ($self->{pid}) {
+    if (!$self->{pid}) {
         print DAEMON_IS_NOT_RUNNING;
         return;
     }
@@ -198,7 +206,29 @@ sub stop {
     $self->before_stop();
     if ($self->{pid}) {
         my $signal = $self->{kill_signal} // POSIX::SIGTERM;
-        $self->{pid}->kill($signal);
+
+        if ($mode eq 'both') {
+            my $new_pid = $self->new_pid;
+            $new_pid->kill($signal) if ($new_pid && $new_pid->is_alive()); # don't use new_pid_is_alive to avoid race cond
+            $self->{pid}->kill($signal);
+        }
+        else { # kill OLD or NEW only when there are NEW and OLD daemons at same time
+            my $new_pid = $self->new_pid;
+            if ($new_pid && $new_pid->is_alive()) { # don't use new_pid_is_alive to avoid race cond
+                if ($mode eq 'new') {
+                    $new_pid->kill($signal)
+                }
+                elsif ($mode eq 'old') {
+                    # we've got OLD daemon pid
+                    # next we've got NEW daemon pid
+                    # it's fixed pid values not
+                    # now if NEW pid is alive, we kill OLD daemon.
+                    # if NEW become OLD in-between, there won't be a race condition
+                    # because we'll just kill non-existing process
+                    $self->{pid}->kill($signal);
+                }
+            }
+        }
     }
 
     $self->after_stop();
@@ -207,6 +237,17 @@ sub stop {
     return 1;
 }
 
+sub stop {
+    shift->_stop('both')
+}
+
+sub stop_new {
+    shift->_stop('new')
+}
+
+sub stop_old {
+    shift->_stop('old')
+}
 
 sub restart {
     my $self = shift;
@@ -221,38 +262,105 @@ sub restart {
     return 1;
 }
 
+# same as start, but
+# starts with --new option
+# expects daemon to be running, sends SIGHUP to it
+# expects new daemon to use pid with ".new" suffix (fails if already running)
+sub _reload {
+    my ($self, $verbose) = @_;
+
+    if (!$self->{_args}{grace_restart_file}) {
+        print GRACE_RESTART_NOT_ALLOWED;
+        return;
+    }
+
+    if (!$self->{pid}) {
+        print DAEMON_IS_NOT_RUNNING;
+        return;
+    } elsif ($self->new_pid_is_alive()) {
+        print GRACE_RESTART_ALREADY_INPROGRESS;
+        return;
+    }
+
+    print RELOADING;
+
+    $self->before_start();
+    # TODO: Add command check
+    my @command = @{$self->{_commands}->{start}};
+
+    if ($self->{_args}->{progress_log}) {
+        # truncate progress_log (better than unlink, because unlink will break 'tail -f')
+        # don't try create-and truncate file here. it will break permissions
+        truncate($self->{_args}->{progress_log}, 0);
+    }
+    unlink $self->{_args}->{grace_restart_file};
+
+    system(
+        @command,
+        '--new',
+        $self->{_args}->{progress_log} ? ('--progress-log' => $self->{_args}->{progress_log}) : ()
+    );
+
+    my $code = $?;
+    $code = $code >> 8;
+    $self->after_start();
+
+    if ($code) {
+        printf NOT_STARTED, $code;
+        return;
+    }
+
+    # what when new daemon will start for sure
+    Time::HiRes::usleep(10_000) while (! $self->new_pid_is_alive());
+
+    # then kill old
+    $self->{pid}->kill(POSIX::SIGHUP);
+
+    print RELOADED;
+
+    if ($verbose && $self->{_args}->{progress_log}) {
+        print "Printing progress log, Ctrl-C to quit tail\n";
+        print "Waiting for log..\n";
+        Time::HiRes::usleep(10_000) while (! -f $self->{_args}->{progress_log});
+        print "Log below:\n";
+        system('tail', '-f', $self->{_args}->{progress_log});
+    }
+}
+
+sub reload {
+    shift->_reload(0)
+}
+
+sub reload_verbose {
+    shift->_reload(1)
+}
 
 sub status {
     my $self = shift;
 
-    unless ($self->{pid}) {
+    if (!$self->{pid}) {
         print DAEMON_IS_NOT_RUNNING;
-        exit 0;
     }
-
-    if ($self->is_alive()) {
-        print DAEMON_ALREADY_RUNNING;
+    elsif ($self->is_alive()) {
+        if ($self->new_pid_is_alive()) {
+            print GRACE_RESTART_ALREADY_INPROGRESS;
+        }
+        else {
+            print DAEMON_ALREADY_RUNNING;
+        }
     }
-
-    exit 0;
 }
 
 
 sub info {
     my $self = shift;
 
-    unless ($self->{pid}) {
-        print DAEMON_IS_NOT_RUNNING;
-        exit 0;
-    }
+    $self->status;
 
     if ($self->is_alive()) {
         my $p = $self->{pid};
-        printf "Daemon is running.\n\tPID: %s\n\tCommand: %s\n\tUser: %s\n",
-            $p->pid(), $p->command(), $p->user();
-    }
-    else {
-        print DAEMON_IS_NOT_RUNNING;
+        printf "\tPID: %s\n\tCommand: %s\n\tUser: %s\n\tProgress-log: %s\n",
+            $p->pid(), $p->command(), $p->user(), $self->{_args}->{progress_log} // 'NONE';
     }
 
     exit 0;
@@ -277,6 +385,18 @@ sub is_alive {
     return $pid->is_alive();
 }
 
+sub new_pid {
+    my $self = shift;
+
+    System::Process::pidinfo(file => $self->{pid_file}.NEW_SUFFIX);
+}
+
+sub new_pid_is_alive {
+    my $self = shift;
+
+    my $new_pid = $self->new_pid;
+    $new_pid && $new_pid->is_alive;
+}
 
 =over
 
